(require kodhy.macros)

(import
  cPickle
  [datetime [datetime]]
  [numpy :as np]
  [pandas :as pd]
  pyper
  [kodhy.util [valcounts pairs seq ret cbind hzscore pds-from-pairs]])

(print "Running")
(setv start-time (datetime.now))

(setv min-cell-size 11) ; Set by a data-use agreement.
(setv sas-data-path-fmt "O:/Desmond/Jennifer/Comorbidities/{}_alldiag.sas7bdat")

(setv comorbs (qw Congestive_heart_failure Cardiac_arrhythmias Valvular_disease Peripheral_vascular_disorders Hypertension__uncomplicated Hypertension__complicated Paralysis Other_neurological_disorders Pulmonary_circulation_disorders Chronic_pulmonary_disease Diabetes__uncomplicated Diabetes__complicated Hypothyroidism Renal_failure Liver_disease Peptic_ulcer_disease Lymphoma Metastatic_cancer Solid_tumor_without_metastasis Rheumatoid_arthritis Coagulopathy Coagulopathy_hemophilia Blood_loss_anemia Deficiency_anemia Obesity Weight_loss Fluid_and_electrolyte_disorders))

(setv orig (dict (rmap [k (qw dual mcaid_only mcare_only)]
  (, k (kwc pd.read-sas :format "sas7bdat" (.format sas-data-path-fmt k))))))

(import __main__)
(unless (hasattr __main__ "_Rproc")
  (setv __main__._Rproc (pyper.R)))
(setv R __main__._Rproc)
(print (.run R "source('//tsclient/luskin/public/assoc_support.R')"))

(setv dfs (dict (rmap [k (qw mcaid_only mcare_only dual)]
  (setv df (get orig k))
  (when (in k (qw mcare_only dual))
    ; Among Medicare beneficiaries, include only those with Part D
    ; coverage for the full year.
    (setv df (ss df (= $partd_mos 12))))
  (setv df (wc df (cbind
    :female (= $mm_sex 2)
    :age_raw $mm_age
    :age_std (hzscore $mm_age)
    :age_std2 (hzscore (** (hzscore $mm_age) 2))
    ; The reference group for race is white.
    :race_black (= $mm_race 2)
    :race_hispanic (= $mm_race 3)
    :race_other_nonwhite (= $mm_race 4)
    :urban (if (= k "mcaid_only") $mcaid_urban $mcare_urban)
    :hiv_docvol_50plus (= $docvol 1)
    :disabled (= $disabled 1)
    (kwc .rename (= (getl (get orig k) : comorbs) 1) :columns (λ (+ "cm_" it)))
    :y_outpatient $otspend
    :y_inpatient $ipspend
    :y_drugs $rxspend
    :y_subtotal (+ $otspend $ipspend $rxspend)
    ;:y_misc $otherspend
    )))
  ; There are relatively few subjects missing on the urban
  ; variable, so just drop them.
  (setv df (kwc .dropna df :subset ["urban"]))
  (setv df (.astype df float))
  (assert (.all (.all (np.isfinite df))))
  (, k df))))
;(setv (get dfs "combined") (pd.concat (rmap [k (qw mcaid_only mcare_only dual)]
;  (setv df (.copy (get dfs k)))
;  (setv old-cols (list df.columns))
;  (setv ($ df no_mcare) (float (= k "mcaid_only")))
;  (setv ($ df no_mcaid) (float (= k "mcare_only")))
;  (getl df : (+ (qw no_mcare no_mcaid) old-cols)))))

(setv results (dict (rmap [k (qw dual mcare_only)]

  (print k)
  (setv df (get dfs k))
  (print df.shape)
  (setv dvs (filt (.startswith it "y_") df.columns))
  (setv ivs (filt (not-in it (+ ["age_raw"] dvs)) df.columns))

  (setv demog (kwc pds-from-pairs (wc df (pairs
    "Female, proportion" (.mean $female)
    "Age, minimum" (.min $age_raw)
    "Age, median" (.median $age_raw)
    "Age, maximum" (.max $age_raw)
    "Race, white, proportion" (.mean (& (= $race_black 0) (= $race_hispanic 0) (= $race_other_nonwhite 0)))
    "Race, black, proportion" (.mean $race_black)
    "Race, Hispanic, proportion" (.mean $race_hispanic)
    "Race, other, proportion" (.mean $race_other_nonwhite)
    "Urban, proportion" (.mean $urban)
    "High-volume HIV care provider, proportion" (.mean $hiv_docvol_50plus)
    "Disabled, proportion" (.mean $disabled)))))

  (setv subtotal-by-n-comorbs (do
    (setv n-comorbs (.astype (kwc .sum :axis 1 (getl df :
      (filt (.startswith it "cm_") df.columns))) int))
    (setv cell-sizes (kwc .sort-index :!ascending (valcounts n-comorbs)))
    (setv cum-cell-sizes (.cumsum cell-sizes))
    (setv max-n-comorbs (min
      (get (ssi cell-sizes (< $ min-cell-size)) -1)
      (first (ssi (.cumsum cell-sizes) (>= $ min-cell-size)))))
    (setv (get n-comorbs (>= n-comorbs max-n-comorbs)) (.format "≥ {}" max-n-comorbs))
    (setv d (.apply
      (.groupby df n-comorbs)
      (λ (kwc pds-from-pairs :dtype object (pairs
        :subjects (len it)
        :median_subtotal (.median ($ it y_subtotal)))))))
    (setv d.index.name "n_comorbs")
    d))

  (.assign R "X.all" (.as-matrix (getl df : ivs)))
  (.run R "X.all = data.frame(X.all)")

  (setv association (kwc pd.concat :axis 1 (rmap [[dv analysis] [
      ["y_subtotal" :quantile-regression-all]
      ["y_outpatient" :quantile-regression-all]
      ["y_inpatient" :logistic-regression-isnonzero]
      ["y_inpatient" :quantile-regression-nonzero]
      ["y_drugs" :quantile-regression-all]]]
    (print k dv (slice analysis 1))
    (.assign R "y" (. (getl df : dv) values))
    (.run R "X = X.all")
    (when (= analysis :quantile-regression-nonzero)
      (.run R "X = X[y != 0,]")
      (.run R "y = y[y != 0]"))
    (when (= analysis :logistic-regression-isnonzero)
      (.run R "y = (y != 0)"))
    (setv r-output (.run R (.format "out = {}(X, y)"
      (if (= analysis :logistic-regression-isnonzero)
        "logistic.reg"
        "quantile.reg"))))
    (unless (= (.count r-output "\n") 1)
      (print r-output))
    (setv colname (+ dv (cond
      [(= analysis :logistic-regression-isnonzero) "_isnonzero"]
      [(= analysis :quantile-regression-nonzero) "_nonzero"]
      [True ""])))
    (setv x (kwc pds-from-pairs :name colname (+
      [
        ["(n)" (.get R "length(y)")]
        ["(central tendency)" (.get R (if (= analysis :logistic-regression-isnonzero)
           "mean(y)"
           "median(y)"))]
        ["(base error)" (.get R (if (= analysis :logistic-regression-isnonzero)
           "mse(y, mean(y))"
           "mae(y, median(y))"))]
        ["(model error)" (.get R "out$error")]]
      (list (zip
        (+ ["Intercept"] ivs)
        (.get R "out$coef"))))))
;    (when (= analysis :logistic-regression-isnonzero)
;    ; Include counterfactual predictions.
;      (setv cpred (kwc pd.Series
;        :index x.index
;        :name (+ x.name "_cpreds")))
;      (setv (getl cpred "(n)") (getl x "(n)"))
;      (setv cpred-R (.get R "out$counterfactual.preds"))
;      (setv (geti cpred (seq (- (len cpred) (len cpred-R)) (- (len cpred) 1)))
;        cpred-R)
;      (setv x (cbind x cpred)))
    x)))  

  (, k {
    "demog" demog
    "association" association
    "subtotal_by_n_comorbs" subtotal-by-n-comorbs}))))

(with [[o (open "//tsclient/luskin/assoc.pkl" "wb")]]
  (cPickle.dump results o))

(print "Took" (- (datetime.now) start-time))
