(require kodhy.macros)

(import
  os.path
  subprocess
  [datetime [datetime]]
  [numpy :as np]
  [pandas :as pd]
  [sklearn.cross-validation :as skcv]
  [sklearn.linear-model :as sklin]
  [sklearn.ensemble :as sken]
  pyper
  [kodhy.util [kfold-cv-pred pds-from-pairs hzscore
    cbind valcounts with-1o-interacts]])

(print "Running")
(setv start-time (datetime.now))

(setv sas-data-path-fmt "O:/Desmond/Jennifer/Comorbidities/all_{}.sas7bdat")

(setv comorbs (qw Congestive_heart_failure Cardiac_arrhythmias Valvular_disease Peripheral_vascular_disorders Hypertension__uncomplicated Hypertension__complicated Paralysis Other_neurological_disorders Pulmonary_circulation_disorders Chronic_pulmonary_disease Diabetes__uncomplicated Diabetes__complicated Hypothyroidism Renal_failure Liver_disease Peptic_ulcer_disease Lymphoma Metastatic_cancer Solid_tumor_without_metastasis Rheumatoid_arthritis Coagulopathy Coagulopathy_hemophilia Blood_loss_anemia Deficiency_anemia Obesity Weight_loss Fluid_and_electrolyte_disorders))

(setv orig (dict (rmap [k (qw dual mcaid_only mcare_only)]
  (, k (kwc pd.read-sas :format "sas7bdat" (.format sas-data-path-fmt k))))))

(import __main__)
(unless (hasattr __main__ "_Rproc")
  (setv __main__._Rproc (pyper.R)))
(setv R __main__._Rproc)
(print (.run R "source('//tsclient/luskin/public/pred_support.R')"))

(defn cv-mae [ivs dv inpatient x-form folds model-f]
  (assert (in x-form [:me :1st :th :inpatient]))
    ; :me for main effects only
    ; :1st for all first-order interactions
    ; :th for some particular theoretically-motived interactions
    ;   suggested by Arleen
    ; :inpatient for :me but also including, as the first
    ;   column, inpatient status
  (when (= x-form :inpatient)
    (setv ivs (cbind (.astype inpatient float) ivs)))
  (when (= x-form :th)
    (setv ivs (kwc pd.concat :axis 1 (+
      [ivs]
      ; Race × gender interactions
      (lc [rc (filt (.startswith it "race_") ivs.columns)]
        (.rename (* ($ ivs female) (getl ivs : rc))
          (+ "female:" rc)))
      ; Age × comorbidity interactions
      (lc [age (qw age_std age_std2) comorb comorbs]
        (.rename (* (getl ivs : age) (getl ivs : (+ "cm_" comorb)))
          (+ age ":cm_" comorb)))))))
  (setv x (.as-matrix ivs))
  (when (= x-form :1st)
    (setv x (with-1o-interacts x)))
  (setv y-true dv.values)
  (setv y-pred (kwc kfold-cv-pred x y-true model-f :folds folds))
  (np.mean (np.abs (- y-true y-pred))))

(defmacro sklearn-model [inner-f]
  `(fn [x-train y-train x-test]
    (setv m ~inner-f)
    (.fit m x-train y-train)
    (.predict m x-test)))

(defmacro sklearn-log-model [inner-f]
  `(fn [x-train y-train x-test]
    (setv m ~inner-f)
    (.fit m x-train (np.log (+ y-train 1)))
    (- (np.exp (.predict m x-test)) 1)))

(defn r-model [r-function]
  (fn [x-train y-train x-test]
    (for [v (qw x_train y_train x_test)]
      (.assign R v (get (locals) v)))
    (print (.run R (.format
      "y_pred = {}(data.frame(x_train), y_train, data.frame(x_test))"
      r-function)))
    (.get R "y_pred")))

(setv models [
  ["Trivial (predict median)" :me (fn [x-train y-train x-test]
    (np.full (second x-test.shape) (np.median y-train)))]
  ["OLS (no log transformation)" :me (sklearn-model
    (sklin.LinearRegression))]
  ["OLS" :me (sklearn-log-model
    (sklin.LinearRegression))]
  ["Ridge regression" :me (sklearn-log-model
    (kwc sklin.RidgeCV :alphas (** 10. (np.arange -5 6))))]
  ["Elastic net" :me (sklearn-log-model
    (kwc sklin.ElasticNetCV :l1-ratio [.01 .05 .1 .25 .5 .75 .9 .95 .99]))]
  ["Elastic net w/ all 1st-order interactions" :1st (sklearn-log-model
    (kwc sklin.ElasticNetCV :l1-ratio [.01 .05 .1 .25 .5 .75 .9 .95 .99]))]
  ["Quantile regression" :me (r-model
    "quantreg.log")]
  ["Quantile regression w/ lasso" :me (r-model
    "quantreg.log.lasso")]
  ["Quantile regression w/ lasso, some interactions" :th (r-model
    "quantreg.log.lasso")]
  ["Quantile regression w/ lasso, inpatient" :inpatient (r-model
    "quantreg.log.lasso.inpatient")]
  ["Random forest" :me (sklearn-log-model
    (kwc sken.RandomForestRegressor :n-estimators 500))]])

(defn theoretical-min-mae [ivs dv cols]
  (setv gb (.groupby (cbind ivs :y dv) cols))
  (np.mean (np.concatenate (rmap [[_ group] gb]
    (.abs (- ($ group y) (.median ($ group y))))))))

(setv results (apply cbind (rmap [k (qw mcaid_only mcare_only dual)]

  (setv ivs (wc (get orig k) (cbind
    :female (= $mm_sex 2)
    :age_std (hzscore $mm_age)
    :age_std2 (hzscore (** (hzscore $mm_age) 2))
    ; The reference group for race is white.
    :race_black (= $mm_race 2)
    :race_hispanic (= $mm_race 3)
    :race_other_nonwhite (= $mm_race 4)
    :disabled (= $disabled 1)
    (kwc .rename (= (getl (get orig k) : comorbs) 1) :columns (λ (+ "cm_" it))))))
  (setv ivs (.astype ivs float))
  (assert (.all (.all (np.isfinite ivs))))

  (setv spent-total ($ (get orig k) totalspend))
  (assert (.all (np.isfinite spent-total)))

  (setv inpatient (wc (get orig k) (|
    (> $medpar_tot_stays 0) (> $nfs 0))))

  (setv folds (list (kwc skcv.KFold (len spent_total)
    :n-folds 10 :+shuffle :random-state 1)))

  (setv results (pd.concat (+
    (rmap [[name x-form f] models]
      (print "Cross-validating:" k name)
      (kwc pd.Series :index [name] (cv-mae ivs spent-total inpatient x-form folds f)))
    [(pds-from-pairs [
      ["[Min MAE, gender and race]" (int (theoretical-min-mae ivs spent-total
        (qw female race_black race_hispanic race_other_nonwhite)))]
      ["[Min MAE, all IVs but age]" (int (theoretical-min-mae ivs spent-total
        (filt (not-in it (qw age_std age_std2)) ivs.columns)))]])])))
  (setv results.name k)
  results)))

(with [[_ (pd.option-context "display.float_format"
    (λ (format it ",.0f")))]]
  (print results))

(print "Took" (- (datetime.now) start-time))
