library(quantreg)

the.lambda = .001
cv.reps = 10

try.pred = function(x.full, y, xtype, twin)
# 'twin' should be a logical variable. If true, we construct
# separate models for Medicare-only and duals.
   {set.seed(5)
    x = xfilter(x.full, xtype)
    mean(replicate(cv.reps,
       {folds = create.folds(x.full$has_medicaid)
        y_pred = crossvalid(x, y, folds = folds, f = function(x_train, y_train, x_test)
            if (xtype == "trivial")
                if (twin)
                    ifelse(x_test$has_medicaid,
                        median(y_train[!! x_train$has_medicaid]),
                        median(y_train[! x_train$has_medicaid]))
                else
                    rep(median(y_train), NROW(x_test))
            else
                if (twin)
                    quantreg.log.lambda.twin(the.lambda, x_train, y_train, x_test)
                else
                    quantreg.log.lambda(the.lambda, x_train, y_train, x_test))
        mean(abs(y - y_pred))}))}

try.pred.2stage = function(x.full, y, xtype.prob, xtype.amount)
   {set.seed(5)
    if (xtype.prob == "trivial")
        try.pred(x.full, y, xtype.amount, twin = F)
    else mean(replicate(cv.reps,
       {folds = create.folds(x.full$has_medicaid)
        y_pred = crossvalid(x.full, y, folds = folds, f = function(x_train, y_train, x_test)
           {lrm = glm(y ~ .,
                cbind(xfilter(x_train, xtype.prob), y = y_train != 0),
                family = binomial)
            if (xtype.amount != "trivial")
                qrm = quantreg::rq(log(y + 1) ~ .,
                    data = subset(cbind(xfilter(x_train, xtype.amount), y = y_train), y != 0),
                    tau = .5,
                    method = "lasso", lambda = the.lambda)
            y_pred = rep(NA_real_, nrow(x_test))
            prob_nonzero_pred = predict(lrm, type = "response",
                newdata = xfilter(x_test, xtype.prob))
            y_pred[prob_nonzero_pred <= .5] = 0
            y_pred[prob_nonzero_pred > .5] = (
                if (xtype.amount == "trivial")
                    median(y_train[y_train != 0])
                else
                    -1 + exp(predict(qrm, newdata =
                        xfilter(x_test, xtype.amount)[prob_nonzero_pred > .5,])))
            y_pred})
        mean(abs(y - y_pred))}))}

xfilter = function(x.full, xtype)
   {if (xtype == "trivial") x.full['has_medicaid'] else
    if (xtype == "demog")   x.full[, -grep('^cm_', colnames(x.full))] else
                            x.full}

crossvalid = function(iv, dv, nfold = 10, folds, f)
# Do cross-validation.
#
# 'iv' and 'dv' can be data frames or matrices (in which case
# each row is considered to be a case) or plain vectors. 'f'
# should have parameters 'train.iv', 'train.dv', and 'test.iv'
# and should return results with the same number of cases as
# 'test.iv' (typically, predictions for test.dv).
#
# Example:
#   x = rnorm(50)
#   y = x + rnorm(50, 0, .5)
#   mean(abs(y - crossvalid(x, y, nfold = 10, f = function(x, y, x2) predict(lm(y ~ x), data.frame(x = x2)))))
   {if (NROW(iv) != NROW(dv))
       stop("iv and dv must have the same number of cases")
    if (missing(folds))
        folds = sample(rep_len(1 : nfold, NROW(iv)))
    else
        nfold = max(folds)
    # Accumulate cross-validated predictions.
    pred = lapply(1 : nfold, function(fold) f(
        select.cases(iv, folds != fold),
        select.cases(dv, folds != fold),
        select.cases(iv, folds == fold)))
    # Return 'pred' in the original order of cases used by
    # 'iv' and 'dv'.
    perm = order(c(recursive = T,
        lapply(1 : nfold, function(fold) which(folds == fold))))
    if (is.null(nrow(pred)[[1]]))
        unlist(pred)[perm]
    else
        do.call(rbind, pred)[perm,]}
select.cases = function(x, i)
    if (is.null(nrow(x))) x[i] else x[i,, drop = F]

create.folds = function(y, k = 10)
# Forked from caret::createFolds (since I had trouble installing
# caret on the server).
   {if (k < length(y))
     {## reset levels so that the possible levels and
      ## the levels in the vector are the same
      y <- factor(as.character(y))
      numInClass <- table(y)
      foldVector <- vector(mode = "integer", length(y))

      ## For each class, balance the fold allocation as far
      ## as possible, then resample the remainder.
      ## The final assignment of folds is also randomized.
      for (i in 1:length(numInClass))
       {## create a vector of integers from 1:k as many times as possible without
        ## going over the number of samples in the class. Note that if the number
        ## of samples in a class is less than k, nothing is producd here.
        min_reps <- numInClass[i] %/% k
        if (min_reps > 0)
         {spares <- numInClass[i] %% k
          seqVector <- rep(1:k, min_reps)
          ## add enough random integers to get  length(seqVector) == numInClass[i]
          if(spares > 0) seqVector <- c(seqVector, sample(1:k, spares))
          ## shuffle the integers for fold assignment and assign to this classes's data
          foldVector[which(y == names(numInClass)[i])] <- sample(seqVector)}
        else
         {## Here there are less records in the class than unique folds so
          ## randomly sprinkle them into folds.
          foldVector[which(y == names(numInClass)[i])] <- sample(1:k, size = numInClass[i])}}}
    else
       {foldVector <- seq(along = y)}

    foldVector}

n.rep.boot = 1000

quantreg.log.lambda = function(lambda, x_train, y_train, x_test, coefci = F)
# Lasso-regularized with a fixed lambda. 
   {model = quantreg::rq(log(y_train + 1) ~ .,
        data = x_train,
        tau = .5,
        method = "lasso", lambda = lambda)
    if (coefci)
       {coefs = model$coefficients
        set.seed(5)
        reps = replicate(n.rep.boot,
           {i = sample.int(length(y_train), length(y_train), rep = T)
            quantreg::rq(log(y_train[i] + 1) ~ .,
                data = x_train[i,],
                tau = .5,
                method = "lasso", lambda = lambda)$coefficients})
        d = data.frame(cbind(coefs,
            t(apply(reps, 1, function(v) quantile(v, c(.025, .975))))))
        colnames(d) = c("estimate", "lo", "hi")
        rownames(d) = names(coefs)
        d}
    else
        exp(predict(model, newdata = x_test)) - 1}

quantreg.log.lambda.twin = function(lambda, x_train, y_train, x_test)
   {train.i = which(!!x_train$has_medicaid)
    test.i = which(!!x_test$has_medicaid)
    preds = rep(NA_real_, nrow(x_test))
    x_train = subset(x_train, select = -has_medicaid)
    x_test = subset(x_test, select = -has_medicaid)
    preds[test.i] = quantreg.log.lambda(lambda,
        x_train[train.i,], y_train[train.i], x_test[test.i,])
    preds[-test.i] = quantreg.log.lambda(lambda,
        x_train[-train.i,], y_train[-train.i], x_test[-test.i,])
    preds}
